"""
-----------------------------------------------------
cfg_lib: utilities to parse standard Virgo .cfg files
-----------------------------------------------------
"""

# B. Swinkels, 2014

from builtins import next
from os import path
import shlex
import logging
from collections import OrderedDict


__all__ = ['parse_config_iter', 'parse_key']

logger = logging.getLogger(__name__)

START_COMMENT = r'/*'
END_COMMENT = r'*/'


def strip_multiline_comments(it):
    """strips /* ... */ comments, while keeping track of line number. Might
    not work with nested comments."""
    enum_lines = enumerate(it, 1)
    for iline, line in enum_lines:
        pos = line.find(START_COMMENT)
        if pos < 0:  # line without comment
            yield iline, line
            continue

        first_part = line[:pos]
        pos += len(START_COMMENT)
        while True:
            pos = line.find(END_COMMENT, pos)
            if pos >= 0:
                break
            # no closing */ found, get next line. raises StopIteration at EOF
            try:
                _, line = next(enum_lines)
            except StopIteration:
                logger.error('Comment started on line %i never closed with "%s"',
                             iline, END_COMMENT)
                yield iline, first_part
                return
            pos = 0
        last_part = line[pos + len(END_COMMENT):]
        yield iline, first_part + last_part


def substitute_alias(iline, fname, frag, alias):
    """substitute in single fragment"""
    for alias_key, alias_value in alias.items():
        # old_frag = frag
        frag = frag.replace(alias_key, alias_value)
        # if frag != old_frag:
        #     logger.debug('*** Substitution from %s to %s on line %i of %s',
        #                  old_frag, frag, iline, fname)
    return frag

def parse_config_iter(fname, alias=None):
    r"""
    Iterate over config file, yielding tuples (fname, iline, split_line)

    This function strips comments, handles include files and correctly splits
    lines that contain quoted strings. Note that mixing or nesting of # and
    /* .. */ comments will likely not work as expected

    Parameters
    ----------
    fname : str
        filename of config file
    alias : dict
        dictionary of constants created with 'define' to be passed recursively
        to include files, should be omitted for user code

    Yields
    ------
    fname : str
        filename of config file (which might not be the same as the original
        fname in case of include files)
    iline : int
        line number
    split_line : list of str
        one line of the config file split in pieces
    """
    fname = path.abspath(fname)
    dirname = path.dirname(fname)

    if alias is None:
        alias = OrderedDict()

    with open(fname) as config_file:
        enum_lines = strip_multiline_comments(config_file)
        for iline, line in enum_lines:
            try:
                # handle line continuation
                while line.endswith('\\\n'):
                    logger.debug('Extending line %d of %s', iline, fname)
                    _, next_line = next(enum_lines)
                    line = line[:-2] + next_line

                # remove #-comments and split correctly on quotes
                split_line = shlex.split(line, True)
                if not split_line:  # skip empty line
                    continue

                # handle define first as special case, should only do
                # substitution in the value, not in the key!
                if split_line[0] == 'define':
                    assert len(split_line) == 3, 'invalid define command'
                    key, val = split_line[1:]
                    val = substitute_alias(iline, fname, val, alias)
                    logger.debug('define %s as %s on line %d of %s:',
                                 key, val, iline, fname)
                    alias[key] = val
                    continue

                # for other lines do substitution all fragments
                split_line = [substitute_alias(iline, fname, frag, alias)
                              for frag in split_line]

                # quick fix for new semicolon added at the end of lines
                if split_line[-1] == ';':
                    split_line = split_line[:-1]

                if split_line[-1].endswith(';'):
                    split_line[-1] = split_line[-1][:-1]

                # recursively handle include files
                if split_line[0] == 'include':
                    assert len(split_line) == 2, 'invalid include'
                    logger.debug('include %s on line %d of %s:',
                                 split_line[1], iline, fname)
                    include_file = path.join(dirname, split_line[1])

                    for fn, il, sl in parse_config_iter(include_file, alias):
                        yield fn, il, sl
                    continue

                yield fname, iline, split_line
            except Exception:
                logger.exception('Exception while parsing line %d of %s:\n%s',
                                 iline, fname, line.strip())
                raise


def parse_key(fname, key):
    """
    returns a list of all lines in the configuration file fname that start with
    the given key. Every line is split considering quotation marks and comments
    and is returned as a list of string, without the keyword itself.

    Parameters
    ----------
    fname : str
        filename of config file
    key : str
        config key

    Examples
    --------

    in somefile.cfg:

    ::

        KEY
        OTHER_KEY
        KEY a
        KEY b c
        # KEY commented
        KEY "with spaces"

    can be parsed like:

    >>> parse_key('somefile.cfg', 'KEY')
    [[], ['a'], ['b', 'c'], ['with spaces']]
    """
    matches = []
    for fname, iline, split_line in parse_config_iter(fname):
        if split_line[0] == key:
            matches.append(split_line[1:])
    return matches


# FIXME: test
def cfg_get_key(fname, key, default=None, nargs=1):
    matches = []
    for fname, iline, split_line in parse_config_iter(fname):
        if split_line[0] == key:
            matches.append(split_line[1:])
    if len(matches) == 0:
        return default
    if len(matches) > 1:
        logger.warning('Found key %s more than once, using first definition', key)
    match = matches[0]
    if len(match) != nargs:
        msg = 'expected %i arguments after key %s, got %i' % (nargs, key, len(match))
        raise ValueError(msg)
    if nargs == 1:
        return match[0]
    else:
        return match
